# Import the typing library so variables can be type cast
from typing import Any, Callable, List, Tuple

# Import the use of abstract classes
from abc import ABC, abstractmethod

# Import the system & get options to process command line arguments
import getopt
import sys

_CLI_Arg_Handled_ = Tuple[Any, ...]
"""Define the return type of the callback handler function"""

_TNGN_ = Callable[[], None]
"""Define a Takes Nothing, Gives Nothing function signature"""

_CLI_Arg_Handler_ = Callable[[List, List, _TNGN_, _TNGN_], _CLI_Arg_Handled_]
"""Define the callback handler function signature"""


class CLIUsage(ABC):
    """Define how a CLI Usage handler should look"""

    @property
    @abstractmethod
    def short_options(self) -> str:
        """Get the command line short version option"""

    @property
    @abstractmethod
    def long_options(self) -> List:
        """Get the command line long version option"""

    @property
    @abstractmethod
    def used(self) -> List:
        """Get a list of all the cli arguments used at runtime"""

    @abstractmethod
    def help(self) -> str:
        """Get the command line usage help"""


class CLArgsProcessor:
    """Processor class to extract the CLI arguments used"""

    def __init__(self, argv, usage: CLIUsage) -> None:
        """Initialize the class instance & set the instance properties"""
        self.__error: bool = False
        try:
            # Extract the options & arguments specified from the supplied arguments
            opts, args = getopt.getopt(argv, usage.short_options, usage.long_options)
        except getopt.GetoptError:
            # Hello there
            self.__error = True
            return
        # Set the options & arguments to process later
        self.__opts = opts
        self.__args = args

    def process(
        self,
        handler: _CLI_Arg_Handler_,
        show_usage_help: _TNGN_,
        show_version: _TNGN_,
    ) -> _CLI_Arg_Handled_:
        """process the command line arguments & return the corresponding data"""
        return handler(self.__opts, self.__args, show_usage_help, show_version)

    @property
    def has_error(self) -> bool:
        """Check if the processor encountered an error"""
        return self.__error


class CLIUsageHandler(CLIUsage):
    """Create a CLI argument usage handler"""

    def __init__(self, short_options: str, long_options: List, help_text) -> None:
        """Set the usage details"""
        self.__short_options: str = short_options
        self.__long_options: List = long_options
        self.arg_processor: CLArgsProcessor = None
        # Initialise the list of CLI arguments used at runtime
        self.__used: List = list()
        # If the help text specified is not a callable function, wrap it in a lambda
        self._help_text: Callable[[], str] = (
            help_text if callable(help_text) else lambda: str(help_text)
        )

    @property
    def short_options(self) -> str:
        """Get all the cli one character options available"""
        return self.__short_options

    @property
    def long_options(self) -> List:
        """Get a list of all the cli long options available"""
        return self.__long_options

    @property
    def used(self) -> List:
        """Get a list of all the cli arguments used at runtime"""
        return self.__used

    def using(self, arg_processor: CLArgsProcessor) -> None:
        """Provide a means of setting the CLI argument processor"""
        self.arg_processor = arg_processor

    def help(self) -> str:
        """Get the usage help for the options"""
        return self._help_text()

    def __str__(self) -> str:
        """Get the object as a human readable string"""
        return "\n".join([self._help_text()])


class SophieLauncher:
    """Define the main CLI script launcher"""

    def __init__(
        self, usage: CLIUsage, app_name: str = __name__, version: str = "0.0.1"
    ) -> None:
        """Initialize the class instance & set the instance properties"""
        self.__error: bool = False
        self.__app_name: str = app_name
        self.__version: str = version
        self.usage: CLIUsage = usage
        # Create a CLI argument processor
        self.arg_processor: CLArgsProcessor = CLArgsProcessor(
            argv=sys.argv[1:], usage=self.usage
        )
        # Check for processor errors
        if self.arg_processor.has_error:
            self.__error = True

    @property
    def has_error(self) -> bool:
        """Check if the launcher encountered an error"""
        return self.__error

    @property
    def app_name(self) -> str:
        """Get the name of the app instance"""
        return self.__app_name

    @property
    def app_version(self) -> str:
        """Get the app instance version"""
        return self.__version

    def show_usage_help(self) -> None:
        """Display the help text by constructing the app name with the usage help"""
        print(self.usage.help(), sep=" ")

    def show_version(self) -> None:
        """Display the version text by constructing the app name with the version"""
        print(self)

    def launch(
        self,
        handler: _CLI_Arg_Handler_,
    ) -> _CLI_Arg_Handled_:
        """Launch the module/script"""
        return self.arg_processor.process(
            handler=handler,
            show_usage_help=self.show_usage_help,
            show_version=self.show_version,
        )

    def __str__(self) -> str:
        """Get the object as a human readable string"""
        return "\n".join([self.app_name, f"Version - {self.app_version}"])


def launcher(
    _usage: CLIUsageHandler,
    app_name: str = __name__,
    version: str = "0.0.1",
    short_options: str = "hvq:",
    long_options: List = ["help", "version", "enquiry="],
    help_text="-q <enquiry>",
) -> Tuple[SophieLauncher, CLIUsageHandler]:
    """Shortcur method to create a SophieLauncher"""
    usage: CLIUsageHandler = _usage(
        short_options=short_options,
        long_options=long_options,
        help_text=help_text,
    )
    sophie: SophieLauncher = SophieLauncher(
        usage=usage,
        app_name=app_name,
        version=version,
    )
    # Check if something went wrong getting a processor
    if sophie.has_error:
        # Display the usage help & GTFO
        sophie.show_usage_help()
        sys.exit(2)
    return sophie, usage


def main() -> None:
    """Main function to run the module"""
    pass


# Make sure the script is being called as a script & not being imported into
# another module file
if __name__ == "__main__":
    # Call the main function
    main()



# Import the typing library so variables can be type cast
from typing import List, Optional, Tuple

# Allow the copying of the dice objects to allow for multiple independant roll values
from copy import copy

# Import the random library
import random

# Define the die face values variable type
_Die_Faces_ = List[int]
# Define the dice cup list variable type
_Dice_Cup_ = List[_Die_Faces_]


def roll_die(_die: _Die_Faces_) -> int:
    """Randomly choose a number from the x sided die face values"""
    return random.choice(_die)


class Dice:
    """Define a Die object instance"""

    def __init__(self, sides: int, name: str = None, val: int = None) -> None:
        """Initial the Die object & set a face value"""
        # Verify that the die will have at least 2 options
        if sides < 2:
            raise ValueError(
                "Number of sides for a valid die needs to be greater than 2"
            )
        # Check if we need to generate a dice name
        if name is None:
            # Genrate the Dn dice name string
            name = f"D{str(sides)}"
        # Create the die face values from 1 to sides +1 because range end is exclusive
        face_values: _Die_Faces_ = [*range(1, sides + 1)]
        # Check if an initial die face value is set
        if val is None:
            # Generate a random starting face value
            val = roll_die(face_values)
        # Make sure the initial die face value is actually within the values on the die
        if val < min(face_values) or val > max(face_values):
            raise ValueError("Dice initial value out of bounds")
        # Set the die instance name
        self.__name: str = name
        # Set the list of die face values
        self.__face_values: _Die_Faces_ = face_values
        # Set the initial starting face value of the die
        self.__rolled: int = val

    @property
    def name(self) -> str:
        """Get the name of the die"""
        return self.__name

    @property
    def faces(self) -> _Die_Faces_:
        """Get the list of face values for the die"""
        return self.__face_values

    @property
    def rolled(self) -> int:
        """Get the rolled face value of the die"""
        return self.__rolled

    def roll(self) -> int:
        """Simulate rolling the die"""
        roll: int = roll_die(self.faces)
        self.__rolled = roll
        return self.rolled

    def __copy__(self):
        """Create a copy of the dice object"""
        return __class__(sides=len(self), name=self.__name, val=self.__rolled)

    def __repr__(self) -> str:
        """Get a string to represent recreating this object"""
        return f"{__class__}(sides={len(self)},name={self.__name},val={self.__rolled})"

    def __str__(self) -> str:
        """Get the object as a human readable string"""
        return f"{self.name} : {self.rolled}"

    def __len__(self) -> int:
        """Return the number of faces the die has"""
        return len(self.__face_values)

    def __eq__(self, o: object) -> bool:
        """Check to see if the dice roll is equal to the operand"""
        return self.__rolled == o

    def __lt__(self, o: object) -> bool:
        """Check to see if the dice roll is less than the operand"""
        return self.__rolled < o

    def __gt__(self, o: object) -> bool:
        """Check to see if the dice roll is greater than the operand"""
        return self.__rolled > o

    def __le__(self, o: object) -> bool:
        """Check to see if the dice roll is less than the operand"""
        return self.__rolled <= o

    def __ge__(self, o: object) -> bool:
        """Check to see if the dice roll is greater than the operand"""
        return self.__rolled >= o


# Define the dice list variable type
_Dice_ = List[Dice]


class DiceShaker:
    """Class to manage rolling multiple dice"""

    def __init__(self, _dice: Optional[_Dice_] = None) -> None:
        """Initial the dice shaker with optional starting dice"""
        # Define the instance properties used, list of dice,
        # current roll total, current min & max roll values
        self.__dice: _Dice_ = list()
        self.__total: int = 0
        self.__min: int = 0
        self.__max: int = 0
        # Check if any dice were supplied at creation
        if _dice is not None:
            # Use the supplied list of dice to populate the shaker
            self.populate(_dice)
            # self.__dice: _Dice_ = _dice
            # self.__total: int = sum([_die.rolled for _die in self.__dice])

    def populate(self, _dice: _Dice_) -> None:
        """Add a subset list of dice to the shaker"""
        for _die in _dice:
            self.add(_die)

    def add(self, _die: Dice, amount: int = 1) -> None:
        """Add a die to dice shaker"""
        # Make sure we are adding a positive number
        if amount < 1:
            raise ValueError("Can't add negative number of die to shaker")
        # I love list comprehension
        # Add a copy of the die to the list to allow for independant roll values
        new_dice: _Dice_ = [copy(_die) for _ in range(1, amount + 1)]
        # Adding the new dice list to existing dice list because thats python
        self.__dice += new_dice
        # Increment the sum total of dice rolls value
        self.__total += sum([_die.rolled for _die in new_dice])
        # Increment the current minimum/maximum roll value by the minimum/maximum
        # roll value of die added
        self.__min += min(_die.faces) * amount
        self.__max += max(_die.faces) * amount

    @property
    def min_roll(self) -> int:
        """Return the minimum value the shaker can roll"""
        return self.__min

    @property
    def max_roll(self) -> int:
        """Return the maximum value the shaker can roll"""
        return self.__max

    @property
    def roll_total(self) -> int:
        """Return the total of the dice rolled"""
        return self.__total

    def empty_shaker(self) -> None:
        """Remove any dice from the shaker list"""
        self.__dice = list()
        self.__total = 0
        self.__min = 0
        self.__max = 0

    def shake(self) -> None:
        """Shake/roll the die in the shaker"""
        # Reset the rolled value
        self.__total = 0
        # Roll the die
        for _die in self.__dice:
            # Add the rolled value to the total
            self.__total += _die.roll()

    def roll(self) -> Tuple[int, Tuple]:
        """Shake & roll the dice to get the results"""
        # Make sure the dice have been shaken
        self.shake()
        # I really love list comprehension lol
        # Get a list of the rolled dice values
        rolled: List = [_die.rolled for _die in self.__dice]
        # Return it as tuple so it's immutable
        return self.roll_total, tuple(rolled)

    def __copy__(self):
        """Create a copy of the dice object"""
        return DiceShaker(self.__dice.copy())

    def __repr__(self) -> str:
        """Get a string to represent recreating this object"""
        return f"{__class__}({repr(self.__dice)})"

    def __str__(self) -> str:
        """Get the object as a human readable string"""
        # TODO work out how to collect the rolled values to add the strings
        return "\n".join(
            [
                f"Rolled : {len(self)} Dice",
                f"Minimum : {self.min_roll}",
                f"Maximum : {self.max_roll}",
                "\n".join([str(d) for d in self.__dice]),
                f"Total : {self.roll_total}",
            ]
        )

    def __len__(self) -> int:
        """Return the number of dice"""
        return len(self.__dice)

    def __getitem__(self, position) -> Dice:
        """Get a specific die from the shaker"""
        return self.__dice[position]

    def __eq__(self, o: object) -> bool:
        """Check to see if the dice roll total is equal to the operand"""
        return self.__total == o

    def __lt__(self, o: object) -> bool:
        """Check to see if the dice roll total is less than the operand"""
        return self.__total < o

    def __gt__(self, o: object) -> bool:
        """Check to see if the dice roll total is greater than the operand"""
        return self.__total > o

    def __le__(self, o: object) -> bool:
        """Check to see if the dice roll total is less than the operand"""
        return self.__total <= o

    def __ge__(self, o: object) -> bool:
        """Check to see if the dice roll total is greater than the operand"""
        return self.__total >= o


def main() -> None:
    """Main function to run the application"""
    import sys

    args = sys.argv[1:]
    print(args)
    die_sides = 15
    d: Dice = Dice(die_sides)
    print(f"Dice(D{die_sides})", d, d < die_sides, d > 0, sep="\n")
    print("Rolls...", d.roll())
    shaker: DiceShaker = DiceShaker([copy(d), copy(d), copy(d)])
    print("New Shaker", shaker, sep="\n")
    shaker.add(d, 2)
    print("Add Dice", shaker, sep="\n")
    shaker.shake()
    print("Shake Dice", shaker, sep="\n")
    print("Roll Shaker", shaker.roll(), shaker, sep="\n")
    shaker.empty_shaker()
    print("Empty Shaker", shaker, sep="\n")
    print("New Shaker Roll", DiceShaker([copy(d), copy(d), copy(d)]).roll(), sep="\n")


# Make sure the script is being called as a script & not being imported into
# another module file
if __name__ == "__main__":
    # Call the main function
    main()



# Import the typing library so variables can be type cast
from typing import List, Tuple

import itertools
import random

# Define the cards & deck variable type structure
_Card_ = Tuple[str, str]
_Deck_ = List[_Card_]
# Define a hand as a collection of cards, like a deck
_Hand_ = List[_Card_]

# Define the number of cards in the deck
STANDARD_DECK_SIZE: int = 52

# Define the card suite types
STANDARD_SUITES: List[str] = ["Hearts", "Clubs", "Diamonds", "Spades"]
# Define the card deck picture/face cards
STANDARD_FACE_CARDS: List[str] = [
    "J",  # Jack
    "Q",  # Queen
    "K",  # King
    "A",  # Ace
]


def shuffle_deck(deck: _Deck_) -> _Deck_:
    """Generate a new shuffled verson of the card deck"""
    # Copy the original so we don't mutate it
    _deck: _Deck_ = deck.copy()
    # Shuffle & return the copy of the deck
    random.shuffle(_deck)
    return _deck


def draw_card(deck: _Deck_, cards: int = 1) -> Tuple[_Hand_, _Deck_]:
    # Verify the drawn number of cards can be satisfied, more than 1 less than deck
    if int(cards) < 0:
        raise ValueError("Invalid number of cards to draw from the deck")
    if cards > len(deck):
        raise ValueError("Insufficient number of cards to draw from the deck")
    # draw the top x number of cards from the start of the deck list
    drawn: _Hand_ = deck[:cards]
    # This reshuffles the deck so is not good for this purpose
    # because it creates a new unordered list
    # We need to just, pop/slice/unshift the cards from the top/staart
    # of the deck list
    deck_remaining: _Deck_ = list(set(set(deck) - set(drawn)))
    return drawn, deck_remaining


def new_deck() -> _Deck_:
    """Create a new deck of cards"""
    # Generate the list of numbers from 2 to 10 & add the picture cards
    ranks: List[_Card_] = [
        str(rank) for rank in list(range(2, 11)) + STANDARD_FACE_CARDS
    ]
    # Create a list of ranked cards combined with suite
    deck: _Deck_ = [card for card in itertools.product(STANDARD_SUITES, ranks)]
    return deck


def new_shuffled_deck() -> _Deck_:
    """Generate a new shuffled deck of cards"""
    return shuffle_deck(new_deck())


def main() -> None:
    """Main function to run the application"""
    deck = new_shuffled_deck()
    for idx, card in enumerate(deck):
        print(f"Card {idx+1} is {card[1]} of {card[0]}")
    drawn, deck_remaining = draw_card(deck, cards=5)
    print(drawn)
    for idx, card in enumerate(deck_remaining):
        print(f"Card {idx+1} is {card[1]} of {card[0]}")


# Make sure the script is being called as a script & not being imported into
# another module file
if __name__ == "__main__":
    # Call the main function
    main()



# Import the typing library so variables can be type cast
from typing import List

# Import the sys for args
import sys

# Import the Sophie launcher system needed
from ewccommons.sophie import (
    _TNGN_,
    _CLI_Arg_Handled_,
    SophieLauncher,
    CLIUsageHandler,
    launcher,
)

# Import the collection of available modules
from ewccommons import dice
from ewccommons import dicerolls
from ewccommons import carddeck

# Define the current library/module version
VERSION: str = "0.0.1.7"


def cli_argument_handler(
    opts: List,
    args: List,
    show_usage_help: _TNGN_,
    show_version: _TNGN_,
) -> _CLI_Arg_Handled_:
    """Handle the processing of the cli arguments"""
    enquiry: str = None
    opt: str
    arg: str
    for opt, arg in opts:
        if opt in ("-h", "--help"):
            # Just show the usage help & halt
            show_version()
            show_usage_help()
            sys.exit()
        elif opt in ("-v", "--version"):
            # Just show the version & halt
            show_version()
            sys.exit()
        elif opt == "--dice":
            # Run the dice main module function
            dice.main()
        elif opt == "--dicerolls":
            # Convert the value to lower, split on space
            # Add any unnamed arguments as potential included option values
            ds: List[str] = arg.lower().split() + args
            # Supply the arguments to the dicerolls main function for handling
            dicerolls.main(*ds)
        elif opt == "--carddeck":
            # Run the carddeck main module function
            carddeck.main()
        elif opt in ("-q", "--enquiry"):
            # Assign the enquiry value
            enquiry = arg
    return enquiry, args


def help_text() -> str:
    """Generate the help text to display"""
    return "\n".join(
        [
            "  -h --help (Show this help)",
            "  -v --version (Show library version)",
            "  --dice (Run the dice module)",
            "  --dicerolls Optional[d6 ...] (Run the dicerolls module)",
            "  --carddeck (Run the carddeck module)",
            "  -q <enquiry> (Make an enquiry)",
        ]
    )


def main() -> None:
    sophie: SophieLauncher
    usage: CLIUsageHandler
    sophie, usage = launcher(
        _usage=CLIUsageHandler,
        app_name="EWC Commons Library",
        version=VERSION,
        short_options="hvq:",
        long_options=["help", "version", "dice", "dicerolls=", "carddeck", "enquiry="],
        help_text=help_text,
    )
    # Let's have a look at what we got
    print(
        "Let's have a look:",
        "#############",
        "Sophie",
        sophie,
        "Usage",
        usage,
        "#############",
        sep="\n",
    )
    # Define the usage & types of variables to be used in this scope
    enquiry: str
    enquiry, args = sophie.launch(cli_argument_handler)
    print(enquiry, args)
    # sophie.show_usage_help()


# Make sure the script is being called as a script & not being imported into
# another module file
if __name__ == "__main__":
    # Call the main function
    main()



# Import the base Dice module to add some convenience
# objects & functions for dice rolling
from ewccommons.dice import _Die_Faces_, Dice

# Define the common die set
D3: Dice = Dice(3, name="D3", val=None)
D4: Dice = Dice(4, name="D4", val=None)
D6: Dice = Dice(6, name="D6", val=None)
D8: Dice = Dice(8, name="D8", val=None)
D10: Dice = Dice(10, name="D10", val=None)
D12: Dice = Dice(12, name="D12", val=None)
D20: Dice = Dice(20, name="D20", val=None)
# Define the common die set face values lists
_D3: _Die_Faces_ = D3.faces
_D4: _Die_Faces_ = D4.faces
_D6: _Die_Faces_ = D6.faces
_D8: _Die_Faces_ = D8.faces
_D10: _Die_Faces_ = D10.faces
_D12: _Die_Faces_ = D12.faces
_D20: _Die_Faces_ = D20.faces


def roll_d3() -> int:
    """Randomly choose a number from the 3 sided die"""
    return D3.roll()


def roll_d4() -> int:
    """Randomly choose a number from the 4 sided die"""
    return D4.roll()


def roll_d6() -> int:
    """Randomly choose a number from the 6 sided die"""
    return D6.roll()


def roll_d8() -> int:
    """Randomly choose a number from the 8 sided die"""
    return D8.roll()


def roll_d10() -> int:
    """Randomly choose a number from the 10 sided die"""
    return D10.roll()


def roll_d12() -> int:
    """Randomly choose a number from the 12 sided die"""
    return D12.roll()


def roll_d20() -> int:
    """Randomly choose a number from the 20 sided die"""
    return D20.roll()


def roll_d100() -> int:
    """Randomly choose a number from the 2 10 sided dice"""
    # Roll a ten sided die as the 10s die value
    # Removing 1 from the tens value allows the multiplier to work
    tens: int = D10.roll() - 1
    # Roll the units & return the sum
    units: int = D10.roll()
    return (tens * 10) + units


def main(*args) -> None:
    """Main function to run the application"""
    # Check if any args were supplied
    if not len(args):
        # Import the sys to get any argv used
        import sys

        # Convert the values to lower case for ease
        args = [str(_).lower() for _ in sys.argv[1:]]
    # Check for die rolls in the supplied arguments list
    if "d3" in args:
        print("D3 Rolls...", roll_d3())
    if "d4" in args:
        print("D4 Rolls...", roll_d4())
    if "d6" in args:
        print("D6 Rolls...", roll_d6())
    if "d8" in args:
        print("D8 Rolls...", roll_d8())
    if "d10" in args:
        print("D10 Rolls...", D10.roll())
    if "d12" in args:
        print("D12 Rolls...", D12.roll())
    if "d20" in args:
        print("D20 Rolls...", D20.roll())
    if "d100" in args:
        print("D100 Rolls...", roll_d100())


# Make sure the script is being called as a script & not being imported into
# another module file
if __name__ == "__main__":
    # Call the main function
    main()
